/*
 * @(#)AirTicketSeller.java	1.00 11/20/01
 *
 * Copyright (c) 2001 by Concurrent Software System Lab (CSSL), Computer Science 
 * Department, The University of Illinois at Chicago (UIC). All Rights Reserved.
 * 
 * CSSL grants you ("Licensee") a non-exclusive, royalty free, license to use,
 * modify and redistribute this software in source and binary code form,
 * provided that i) this copyright notice and license appear on all copies of
 * the software; and ii) Licensee does not utilize the software in a manner
 * which is disparaging to CSSL.
 * 
 * This software is provided "AS IS," without a warranty of any kind. ALL
 * EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES, INCLUDING ANY
 * IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE OR
 * NON-INFRINGEMENT, ARE HEREBY EXCLUDED. CSSL AND ITS LICENSORS SHALL NOT BE
 * LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING
 * OR DISTRIBUTING THE SOFTWARE OR ITS DERIVATIVES. IN NO EVENT WILL CSSL OR ITS
 * LICENSORS BE LIABLE FOR ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT,
 * INDIRECT, SPECIAL, CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER
 * CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF THE USE OF
 * OR INABILITY TO USE SOFTWARE, EVEN IF CSSL HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES.
 * 
 * This software is not designed or intended for use in on-line control of
 * aircraft, air traffic, aircraft navigation or aircraft communications; or in
 * the design, construction, operation or maintenance of any nuclear
 * facility. Licensee represents and warrants that it will not use or
 * redistribute the Software for such purposes.
 */


package agent.airTicketSeller;

import agent.generic.GSP;
import agent.generic.Message;
import agent.generic.MSP;
import agent.generic.Agent;
import agent.generic.Goal;
import agent.generic.Plan;
import agent.seller.Seller;

import net.jini.core.entry.Entry;
import net.jini.core.lookup.ServiceID;
import net.jini.core.event.*;

import net.jini.lookup.entry.ServiceInfo;
import net.jini.lookup.entry.Name;

import java.rmi.RemoteException;

import java.util.Enumeration;
import java.util.Hashtable;
import java.lang.reflect.Array;
import java.net.URL;
import java.io.*;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.tree.*;
import javax.swing.border.*;


public class AirTicketSeller extends Seller
{
    private static final String PRODUCT
                                = "Seller_Agent";
    private static final String MANUFACTURER = "CSSL@UIC";
    private static final String VENDOR = MANUFACTURER;
    private static final String VERSION = "Jini 1.1";

    private static final String REMOTE_AGENT = "Buyer";

    private static boolean inTransaction = false;


    // ----------------------------------- 
    // variables defined in knowledge-base
    // -----------------------------------
    ActionListener menuActionListener;
    ActionListener buttonActionListener;
    JFrame frameKnowledge; 
  
    private String fromCity;
    private String departureTime;
    private String toCity;
    private String arrivalTime;
    private String dateFly;
    private String airlineFly;
    private float startingPrice;
    private float lowestPrice;
    private int soldTickets;
    private int availableTickets;

    private JTextField inputFrom = new JTextField("Chicago");
    private JTextField inputTo = new JTextField("Dayton");
    private JTextField inputDepartureTime = new JTextField("12:00pm");
    private JTextField inputArrivalTime = new JTextField("12:00pm");
    private JTextField inputDate = new JTextField("12/05/01");
    private JTextField inputAirline = new JTextField("United Airline");
    private JTextField inputStartingPrice = new JTextField("300.00");
    private JTextField inputLowestPrice = new JTextField("200.00"); 
    private JTextField inputSoldTickets = new JTextField("0");
    private JTextField inputAvailableTickets = new JTextField("15"); 


    // -------------------------------- 
    // variables defined in plan module
    // --------------------------------
    
    private PlanToBeFoundByBuyer plan_BeFoundByBuyer = new PlanToBeFoundByBuyer();
    private PlanToFindBuyer plan_FindBuyer = new PlanToFindBuyer();
    private PlanToBeAskedPrice plan_BeAskedPrice = new PlanToBeAskedPrice();
    private PlanToBeNegotiatedPrice plan_BeNegotiatedPrice = new PlanToBeNegotiatedPrice();

   
    // -------------------------------- 
    // variables defined in goal module
    // --------------------------------

    private ActionListener menuActionListener1;
    private ActionListener buttonActionListener1;
    private JFrame goalFrame; 
    private JTree goalTree;
    private JEditorPane goalHtmlPane;

    private static final String GOAL_FILE = "goal.txt";
    private MyGoal myGoal = new MyGoal(); // must be defined after plans


    // ===========
    // Constructor
    // ===========

    public AirTicketSeller() throws RemoteException {
        super();
        myAgentName = "AirTicketSeller";
        myPackageName = "airTicketSeller";
        serviceAttrs = new Entry[]{new ServiceInfo(PRODUCT,
                                                   MANUFACTURER,
                                                   VENDOR,
                                                   VERSION,
                                                   null, null),
                                   new Name("Seller")};
        serviceOnline = true;
    }


    // =============
    // Main Function
    // =============

    public static void main(String[] args) {

        // ---------------------------------
        // Create the AirTicketSeller object
        // ---------------------------------

        try {
            thisClass = new AirTicketSeller();  // ** VERY IMPORTANT !!! **
        } catch (RemoteException e) {
            System.out.println("RemoteException while exporting this object "
                               + e.toString());
            System.exit(-1);
        }

        ((AirTicketSeller)thisClass).initAgent(args, "AirTicketSeller agent console");
       
        // initialize the knowledge-base
        console.appendInfo("Initialize the knowledge-base ...");
        ((AirTicketSeller)thisClass).saveKnowledge();
    
        console.appendInfo("Press Ctrl-R to start autonomousRun() ...");

        // ((AirTicketSeller)thisClass).autonomousRun(); // redefined in this class

    }

    
    /********************
     *                  *
     *    Goal Class    *
     *                  *
     ********************/

    private class MyGoal extends Goal {

        public MyGoal() {
            setGoalName("sell air ticket");
            subGoal = new Goal[1];
            plan = null;

            // subGoal[0] -- find buyer
            subGoal[0] = new Goal();
            subGoal[0].goalName = "find buyer"; 
            subGoal[0].subGoal = null;
            subGoal[0].plan = new Plan[2];
            subGoal[0].plan[0] = plan_FindBuyer;
            subGoal[0].plan[1] = plan_BeFoundByBuyer;
       }         
    }

    private boolean achieveGoal(Goal goal) {
     
        if (goal == null) return false;

        if (goal.subGoal != null) {
            for (int i=0; i<Array.getLength(goal.subGoal); i++) {
                if (goal.subGoal[i] != null) {
                    System.out.println("goal = " + goal.subGoal[i].goalName);             
                    achieveGoal(goal.subGoal[i]);  
                }  
            }
        }    

        // achieve this goal, invoke associated plan or plans
        System.out.println("achieve this goal = " + goal.goalName);
        if (goal.plan != null && goal.plan[1] != null) {
            goal.plan[1].start();     // wait for messages coming
        }
        
        return true;
    }
   
 
    
    /********************
     *                  *
     *   Plan Classes   *
     *                  *
     ********************/


    private class PlanToBeFoundByBuyer extends Plan {
        
        // override the method start() defined in class Plan
        public void start() {
            waitForQueryFromBuyer();
        }

        // action 1: wait for query from buyer
        public void waitForQueryFromBuyer() {
            // doing nothing and wait ...
        }
        
        // action 2: reply yes or no
        public Message sellerYesOrNo(Message message) {
            Message outgoingMessage = null;     
       
            // analyze the message and generate the outgoing message 

            if (message.getContent().indexOf("Air") != -1 || 
                message.getContent().indexOf("air") != -1) {
                if (getMyAgentName().indexOf("Air") != -1 ||
                    getMyAgentName().indexOf("air") != -1) { 
                    outgoingMessage = new Message(thisServiceID, 
                                     "Yes, I do. I am an air ticket seller.", 
                                      Message.SELLER_YES); 
                } else {
                    outgoingMessage = new Message(thisServiceID, 
                                     "No, I don't. I am NOT an air ticket seller.", 
                                      Message.SELLER_NO); 
                }
            } else if (message.getContent().indexOf("Train") != -1 || 
                message.getContent().indexOf("train") != -1) {
                if (getMyAgentName().indexOf("Train") != -1 ||
                    getMyAgentName().indexOf("train") != -1) { 
                    outgoingMessage = new Message(thisServiceID, 
                                     "Yes, I do. I am an train ticket seller.", 
                                      Message.SELLER_YES); 
                } else {
                    outgoingMessage = new Message(thisServiceID, 
                                     "No, I don't. I am NOT an train ticket seller.", 
                                      Message.SELLER_NO); 
                }
            } 
                 
            outgoingMessage.setReceiverID(message.getSenderID());     
     
            return outgoingMessage;                  
        }         
       
        // action 3: reply if the Depart and Arrive place ok
        public Message placeYesOrNo(Message message) {
            Message outgoingMessage = null;     
       
            // analyze the message and generate the outgoing message 

            if (message.varString[0].equals(fromCity) &&
                message.varString[1].equals(toCity)) {
                outgoingMessage = new Message(thisServiceID, 
                                      "Yes, I sell tickets from " 
                                      + message.varString[0]
                                      + " to " + message.varString[1] + ".", 
                                      Message.PLACE_YES); 
            } else {
                outgoingMessage = new Message(thisServiceID, 
                                      "No, I do NOT have tickets from " 
                                      + message.varString[0] + " to " 
                                      + message.varString[1] +".", 
                                      Message.PLACE_NO); 
            }
            
            outgoingMessage.setReceiverID(message.getSenderID());     
     
            return outgoingMessage;                  
        }
        
        // action 4: reply if the Date and NumOfTickets ok
        public Message dateAndNumOfTicketsYesOrNo(Message message) {
            Message outgoingMessage = null;     
            String ticketStr = null;        

            // analyze the message and generate the outgoing message 

            if (message.varString[0].equals(dateFly) && availableTickets > 0) {
                if (message.varInt[0] <= availableTickets) {
                    if (message.varInt[0] > 1) {
                        ticketStr = "tickets";
                    } else {
                        ticketStr = "ticket"; 
                    }   
                    outgoingMessage = new Message(thisServiceID, 
                                      "Ok, I have " + message.varInt[0]
                                      + " air " + ticketStr + " on " 
                                      + message.varString[0] + ".", 
                                      Message.DATE_NUMOFTICKETS_YES);
                } else {
                    if (availableTickets > 1) {
                        ticketStr = "tickets";
                    } else {
                        ticketStr = "ticket"; 
                    } 
                    outgoingMessage = new Message(thisServiceID, 
                                      "But I only have " + availableTickets
                                      + " air " + ticketStr + " on " 
                                      + message.varString[0] + ".", 
                                      Message.DATE_NUMOFTICKETS_YES);
                } 
            } else {
                outgoingMessage = new Message(thisServiceID, 
                                      "Sorry, I do NOT have any tickets on" 
                                      + message.varString[0] + ".",
                                      Message.DATE_NUMOFTICKETS_NO); 
            }
            
            outgoingMessage.setReceiverID(message.getSenderID());     
     
            return outgoingMessage;                  
        }
    }

   
    private class PlanToFindBuyer extends Plan {

        // override the method start() defined in class Plan
        public void start() {
            makeQueries();
        }

        // action 1: make the queries
        public void makeQueries() {

            if (isEmptyServiceTable()) return;            

            Agent remoteAgent = null;
            int nServices = serviceTable.size();

            if (nServices > 0) {

                for(Enumeration e=serviceTable.elements(); e.hasMoreElements();) {
                    remoteAgent = (Agent)e.nextElement();
                    Message outgoingMessage = new Message(thisServiceID,
                                              "Do you buy air tickets?",
                                              Message.FIND_BUYER);
                    outgoingMessage.setReceiverID(remoteAgent.getServiceID());
                    dispatchOutgoingMessage(outgoingMessage);
                }
            }
        }

        // action 2: wait for replies from the buyers

    }
    
    private class PlanToBeAskedPrice extends Plan {

        // override the method start() defined in class Plan
        public void start() {
            waitForQueryFromBuyer();
        }

        // action 1: wait for query from buyer
        public void waitForQueryFromBuyer() {
            // doing nothing and wait ...
        }
        
        // action 2: reply with price
        public Message replyWithPrice(Message message) {
            Message outgoingMessage = null;     
       
            // analyze the message and generate the outgoing message
                    
            outgoingMessage = new Message(thisServiceID, 
                                  "The air ticket costs $" + startingPrice +".",
                                  Message.REPLY_PRICE); 
            outgoingMessage.varFloat = new float[1];
            outgoingMessage.varFloat[0] = startingPrice;
            outgoingMessage.setReceiverID(message.getSenderID());     
     
            return outgoingMessage;                   
        } 
    }

    private class PlanToBeNegotiatedPrice extends Plan {

        // override the method start() defined in class Plan
        public void start() {
            waitForQueryFromBuyer();
        }

        // action 1: wait for query from buyer
        public void waitForQueryFromBuyer() {
            // doing nothing and wait ...
        }
        
        // action 2: reply with lower price
        public Message replyWithLowerPrice(Message message) {
            Message outgoingMessage = null;     
       
            // analyze the message and generate the outgoing message
                    
            outgoingMessage = new Message(thisServiceID, 
                                  "The lowest price is $" + lowestPrice + ".", 
                                  Message.REPLY_LOWER_PRICE); 

            outgoingMessage.varFloat = new float[1];
            outgoingMessage.varFloat[0] = lowestPrice;
            outgoingMessage.varInt = new int[1];
            outgoingMessage.varInt[0] = availableTickets;                                 
            outgoingMessage.setReceiverID(message.getSenderID());     
     
            return outgoingMessage;                   
        } 
    }  


    /*************************
     *                       *
     *  Internal Structure   *
     *                       *
     *************************/

    ////////////////////////////
    // Incoming Message Section
    ////////////////////////////

    // ------------------- 
    // welcome from seller
    // -------------------

    protected void MPU_In_WelcomeFromBuyer(Message message) {
        Agent remoteAgent = getRemoteAgent_In(message);
        String remoteAgentName = remoteAgent.getAgentName();

        console.appendInfo(remoteAgentName + ": " +  message.getContent());

        // asynchronous message passing
        MSP.sendInternal(this, message);
    }

    // ------------------- 
    // plan_BeFoundByBuyer
    // -------------------

    protected void MPU_In_FindSeller(Message message) {
        Agent remoteAgent = getRemoteAgent_In(message);
        String remoteAgentName = remoteAgent.getAgentName();

        console.appendInfo(remoteAgentName + ": " +  message.getContent());

        // asynchronous message passing
        MSP.sendInternal(this, message);
    }

    protected void MPU_In_AskDepartAndArrivePlace(Message message) {
        Agent remoteAgent = getRemoteAgent_In(message);
        String remoteAgentName = remoteAgent.getAgentName();

        console.appendInfo(remoteAgentName + ": " +  message.getContent());

        // asynchronous message passing
        MSP.sendInternal(this, message);
    }

    protected void MPU_In_AskDateAndNumOfTickets(Message message) {
        Agent remoteAgent = getRemoteAgent_In(message);
        String remoteAgentName = remoteAgent.getAgentName();

        while (inTransaction) {
            try {Thread.sleep(3000);} catch (InterruptedException e) { }
        }
        inTransaction = true;

        console.appendInfo(remoteAgentName + ": " +  message.getContent());

        // asynchronous message passing
        MSP.sendInternal(this, message);
    }


    // -------------- 
    // plan_FindBuyer
    // --------------

    protected void MPU_In_BuyerYesOrNo(Message message) {
        Agent remoteAgent = getRemoteAgent_In(message);
        String remoteAgentName = remoteAgent.getAgentName();

        console.appendInfo(remoteAgentName + ": " +  message.getContent());

        // ***** delete buyer agent here !!!
        if (message.getType() == Message.SELLER_NO) {
            deleteBuyer_In(message); 
        }       

        // asynchronous message passing
        MSP.sendInternal(this, message);
    }

    // ----------------- 
    // plan_BeAskedPrice
    // -----------------

    protected void MPU_In_BeAskedPrice(Message message) {
        Agent remoteAgent = getRemoteAgent_In(message);
        String remoteAgentName = remoteAgent.getAgentName();
        
        console.appendInfo(remoteAgentName + ": " +  message.getContent());

        // asynchronous message passing
        MSP.sendInternal(this, message);
    }

    // ---------------------- 
    // plan_BeNegotiatedPrice
    // ----------------------

    protected void MPU_In_BeNegotiatedPrice(Message message) {
        Agent remoteAgent = getRemoteAgent_In(message);
        String remoteAgentName = remoteAgent.getAgentName();

        console.appendInfo(remoteAgentName + ": " +  message.getContent());

        // asynchronous message passing
        MSP.sendInternal(this, message);
    }

    // -------------------- 
    // plan_BeComparedPrice
    // --------------------

    protected void MPU_In_BeComparedPrice(Message message) {
        Agent remoteAgent = getRemoteAgent_In(message);
        String remoteAgentName = remoteAgent.getAgentName();
        
        console.appendInfo(remoteAgentName + ": " +  message.getContent());

        // asynchronous message passing
        MSP.sendInternal(this, message);
    }

    // --------------------- 
    // plan_BeConfirmedPrice
    // ---------------------

    protected void MPU_In_BeConfirmedPrice(Message message) {
        Agent remoteAgent = getRemoteAgent_In(message);
        String remoteAgentName = remoteAgent.getAgentName();

        soldTickets = soldTickets + message.varInt[0];
        inputSoldTickets.setText(Integer.toString(soldTickets));
        availableTickets = availableTickets - message.varInt[0];
        inputAvailableTickets.setText(Integer.toString(availableTickets));
        console.appendInfo(remoteAgentName + ": " +  message.getContent());

        inTransaction = false;

        // asynchronous message passing
        MSP.sendInternal(this, message);
    }


    ////////////////////////////
    // Outgoing Message Section
    ////////////////////////////

    // ---------------- 
    // welcome to buyer
    // ----------------

    protected void MPU_Out_WelcomeToBuyer(Message outgoingMessage) {
        Agent remoteAgent = getRemoteAgent_Out(outgoingMessage);
        String remoteAgentName = remoteAgent.getAgentName();

        console.appendInfo(getAgentName() + " (to "
                                          + remoteAgent.getAgentName()
                                          + "): " + outgoingMessage.getContent());
        // asynchronous message passing
        MSP.sendExternal((GSP)remoteAgent.getService(), outgoingMessage);
    }

 
    // ------------------- 
    // plan_BeFoundByBuyer
    // -------------------
 
    protected void MPU_Out_SellerYesOrNo(Message outgoingMessage) {
        Agent remoteAgent = getRemoteAgent_Out(outgoingMessage);
        String remoteAgentName = remoteAgent.getAgentName();

        console.appendInfo(getAgentName() + " (to " + remoteAgentName
                                    + "): " + outgoingMessage.getContent());

        // ***** delete buyer agent here !!!
        if (outgoingMessage.getType() == Message.SELLER_NO) {
            deleteBuyer_Out(outgoingMessage); 
        }       
 
        // asynchronous message passing
        MSP.sendExternal((GSP)remoteAgent.getService(), outgoingMessage);
    }

    protected void MPU_Out_PlaceYesOrNo(Message outgoingMessage) {
        Agent remoteAgent = getRemoteAgent_Out(outgoingMessage);
        String remoteAgentName = remoteAgent.getAgentName();

        console.appendInfo(getAgentName() + " (to " + remoteAgentName
                                    + "): " + outgoingMessage.getContent());

        // ***** delete buyer agent here !!!
        if (outgoingMessage.getType() == Message.PLACE_NO) {
            deleteBuyer_Out(outgoingMessage); 
        }       
 
        // asynchronous message passing
        MSP.sendExternal((GSP)remoteAgent.getService(), outgoingMessage);
    }

    protected void MPU_Out_DateAndNumOfTicketsYesOrNo(Message outgoingMessage) {
        Agent remoteAgent = getRemoteAgent_Out(outgoingMessage);
        String remoteAgentName = remoteAgent.getAgentName();

        console.appendInfo(getAgentName() + " (to " + remoteAgentName
                                    + "): " + outgoingMessage.getContent());

        // ***** delete buyer agent here !!!
        if (outgoingMessage.getType() == Message.DATE_NUMOFTICKETS_NO) {
            deleteBuyer_Out(outgoingMessage); 
        }       
 
        // asynchronous message passing
        MSP.sendExternal((GSP)remoteAgent.getService(), outgoingMessage);
    }


    // -------------- 
    // plan_FindBuyer
    // --------------

    protected void MPU_Out_FindBuyer(Message outgoingMessage) {
        Agent remoteAgent = getRemoteAgent_Out(outgoingMessage);
        String remoteAgentName = remoteAgent.getAgentName();

        console.appendInfo(getAgentName() + " (to "
                                          + remoteAgent.getAgentName()
                                          + "): " + outgoingMessage.getContent());
        // asynchronous message passing
        MSP.sendExternal((GSP)remoteAgent.getService(), outgoingMessage);
    }

    // ----------------- 
    // plan_BeAskedPrice
    // -----------------

    protected void MPU_Out_ReplyWithPrice(Message outgoingMessage) {
        Agent remoteAgent = getRemoteAgent_Out(outgoingMessage);
        String remoteAgentName = remoteAgent.getAgentName();

        console.appendInfo(getAgentName() + " (to "
                                          + remoteAgent.getAgentName()
                                          + "): " + outgoingMessage.getContent());
        // asynchronous message passing
        MSP.sendExternal((GSP)remoteAgent.getService(), outgoingMessage);
    }

    // ---------------------- 
    // plan_BeNegotiatedPrice
    // ----------------------

    protected void MPU_Out_ReplyWithLowerPrice(Message outgoingMessage) {
        Agent remoteAgent = getRemoteAgent_Out(outgoingMessage);
        String remoteAgentName = remoteAgent.getAgentName();

        console.appendInfo(getAgentName() + " (to "
                                          + remoteAgent.getAgentName()
                                          + "): " + outgoingMessage.getContent());
        // asynchronous message passing
        MSP.sendExternal((GSP)remoteAgent.getService(), outgoingMessage);
    }


  
    //////////////////////////
    // Utility Method Section
    //////////////////////////

    public void autonomousRun() {
         
        // achieve goals
        // ((PlanToFindBuyer)(myGoal.subGoal[0].plan[0])).makeQueries();
        
        achieveGoal(myGoal);

        // try {Thread.sleep(3000);} catch (InterruptedException e) { }
        // console.appendInfo("Now I am waiting for airticket buyers to talk to me ...");

        stayUpForLookupSrvcAndInvoke();
        console.appendInfo("\nTime is up ... Terminate the JoinManager ...");
        joinMgr.terminate();
        console.appendInfo("\nJoin Manager Terminated ... exiting ...\n");
        System.exit(0);
    }

    protected void invokePlan(RemoteEvent ev) {

        Hashtable serviceTableTmp = (Hashtable) serviceTable.clone();
        updateServices();
 
        Agent remoteAgent = null;
        Agent remoteAgent1 = null; 
        int nServices = serviceTable.size();
        int nnServices = serviceTableTmp.size();

        // System.out.println("serviceTable = " + nServices);
        // System.out.println("serviceTableTmp = " + nnServices);

        if (nServices > 0) {
            boolean newAgent;
            for(Enumeration e=serviceTable.elements(); e.hasMoreElements();) {
                remoteAgent = (Agent)e.nextElement();
                newAgent = true;
                for (Enumeration e1=serviceTableTmp.elements(); e1.hasMoreElements();) {
                     remoteAgent1 = (Agent)e1.nextElement();
                     if (remoteAgent.getAgentName().equals(remoteAgent1.getAgentName())) {
                         newAgent= false;
                         break; 
                     }
                }                
                if (newAgent) {
                    // Message outgoingMessage = new Message(thisServiceID,
                    //                           "Do you buy air tickets?",
                    //                           Message.FIND_BUYER);
                    Message outgoingMessage = new Message(thisServiceID,
                                              "Welcome to the Jini Community !",
                                              Message.WELCOME);
                    outgoingMessage.setReceiverID(remoteAgent.getServiceID());
                    dispatchOutgoingMessage(outgoingMessage);
                }
            }
        }
    }


    protected void dispatchIncomingMessage(Message message) {

        if (isEmptyServiceTable()) return;                

        switch (message.getType()) {
            // welcome message from buyer
            case Message.WELCOME:
                MPU_In_WelcomeFromBuyer(message); 
                break;         
            // ------------------- 
            // plan_BeFoundByBuyer
            // -------------------
            case Message.FIND_SELLER:
                MPU_In_FindSeller(message);
                break;
            case Message.DEPART_ARRIVE_PLACE:
                MPU_In_AskDepartAndArrivePlace(message);
                break;
            case Message.DATE_NUMOFTICKETS:
                MPU_In_AskDateAndNumOfTickets(message);
                break;
            // -------------- 
            // plan_FindBuyer
            // -------------- 
            case Message.BUYER_YES:
                MPU_In_BuyerYesOrNo(message);
                break;
            case Message.BUYER_NO:
                MPU_In_BuyerYesOrNo(message);
                break;
            // ----------------- 
            // plan_BeAskedPrice
            // ----------------- 
            case Message.ASK_PRICE:
                MPU_In_BeAskedPrice(message);
                break;
            // ---------------------- 
            // plan_BeNegotiatedPrice
            // ---------------------- 
            case Message.NEGOTIATE_PRICE:
                MPU_In_BeNegotiatedPrice(message);
                break;
            // -------------------- 
            // plan_BeComparedPrice
            // -------------------- 
            case Message.COMPARE_PRICE:
                MPU_In_BeComparedPrice(message);
                break;
            // --------------------- 
            // plan_BeConfirmedPrice
            // --------------------- 
            case Message.CONFIRM_PRICE:
                MPU_In_BeConfirmedPrice(message);
                break;

            default: super.dispatchIncomingMessage(message);
        }
    }

    protected Message makeDecision(Message message) {

        Message outgoingMessage = null;
        
        switch (message.getType()) {
            // welcome message from seller
            case Message.WELCOME:
                // doing nothing: outgoingMessage = null
                break;                     
            // -------------------
            // plan_BeFoundByBuyer
            // -------------------
            case Message.FIND_SELLER:
                outgoingMessage = plan_BeFoundByBuyer.sellerYesOrNo(message);
                break;
            case Message.DEPART_ARRIVE_PLACE:
                outgoingMessage = plan_BeFoundByBuyer.placeYesOrNo(message);
                break;
            case Message.DATE_NUMOFTICKETS:
                outgoingMessage = plan_BeFoundByBuyer.dateAndNumOfTicketsYesOrNo(message);
                break;
            // --------------
            // plan_FindBuyer
            // -------------- 
            case Message.BUYER_YES:
                // do nothing
                break;
            case Message.BUYER_NO:
                // add into dServiceTable and update serviceTable
                // plan_FindBuyer.deleteBuyer(message); delete in MPU_In_XXX
                // doing nothing here
                break;
            // -----------------
            // plan_BeAskedPrice
            // -----------------
            case Message.ASK_PRICE:
                outgoingMessage = plan_BeAskedPrice.replyWithPrice(message);
                break;
            // ----------------------
            // plan_BeNegotiatedPrice
            // ----------------------
            case Message.NEGOTIATE_PRICE:
                outgoingMessage = plan_BeNegotiatedPrice.replyWithLowerPrice(message);
                break;
            // --------------------
            // plan_BeComparedPrice
            // --------------------
            case Message.COMPARE_PRICE:
                // do nothing
                break;
            // ---------------------
            // plan_BeConfirmedPrice
            // ---------------------
            case Message.CONFIRM_PRICE:
                // do nothing
                break;
            
            default:
                outgoingMessage = super.makeDecision(message);    
                break; 
        }
         
        return outgoingMessage;
    }


    protected void dispatchOutgoingMessage(Message outgoingMessage) {

        if (outgoingMessage == null) {
            System.out.println("** No outgoing message generated!");
            return;
        }      

        if (isEmptyServiceTable()) return;       

        switch (outgoingMessage.getType()) {
            // welcome message to buyer
            case Message.WELCOME:
                MPU_Out_WelcomeToBuyer(outgoingMessage); 
                break;          
            // ------------------- 
            // plan_BeFoundByBuyer
            // -------------------
            case Message.SELLER_YES:
                MPU_Out_SellerYesOrNo(outgoingMessage);
                break;
            case Message.SELLER_NO:
                MPU_Out_SellerYesOrNo(outgoingMessage);
                break;
            case Message.PLACE_YES:
                MPU_Out_PlaceYesOrNo(outgoingMessage);
                break;
            case Message.PLACE_NO:
                MPU_Out_PlaceYesOrNo(outgoingMessage);
                break;  
            case Message.DATE_NUMOFTICKETS_YES:
                MPU_Out_DateAndNumOfTicketsYesOrNo(outgoingMessage);
                break;
            case Message.DATE_NUMOFTICKETS_NO:
                MPU_Out_DateAndNumOfTicketsYesOrNo(outgoingMessage);
                break;   
            // --------------
            // plan_FindBuyer
            // -------------- 
            case Message.FIND_BUYER:
                MPU_Out_FindBuyer(outgoingMessage);
                break;
            // -----------------
            // plan_BeAskedPrice
            // -----------------
            case Message.REPLY_PRICE:
                MPU_Out_ReplyWithPrice(outgoingMessage);
                break;
            // -----------------
            // plan_BeNegotiatedPrice
            // -----------------
            case Message.REPLY_LOWER_PRICE:
                MPU_Out_ReplyWithLowerPrice(outgoingMessage);
                break;

            default: super.dispatchOutgoingMessage(outgoingMessage);
        }
    }


    /*************************
     *                       *
     *   My User Interface   * 
     *                       *
     *************************/ 

    ///////////////// 
    // Goal Module //
    /////////////////

    public void editGoal() {
 
        goalFrame = new JFrame("AirTicketSeller agent goal and plan module");
        buttonActionListener1 = new ButtonActionListener();
        JMenuBar menuBar = new JMenuBar();
     
        EmptyBorder border5 = new EmptyBorder(5,5,5,5);
        EmptyBorder border10 = new EmptyBorder(10,10,10,10);

        JLabel labelMain = new JLabel("AirTicketSeller (ID: " 
                                                   + getAgentName() + ")");

        Container pane = goalFrame.getContentPane();
        pane.setLayout(new BorderLayout());
       
        // top panel
        JPanel top = new JPanel();
	  top.setLayout(new BorderLayout());
	  pane.add(top, BorderLayout.NORTH);

	  menuBar = createMenuBar();
	  top.add(menuBar, BorderLayout.NORTH);

        JPanel labelPanel = new JPanel();
	  labelPanel.setLayout(new BorderLayout());
        labelPanel.setBorder(border5);
        ImageIcon image = new ImageIcon("..\\images\\seller.gif",  "Seller.icon");
        labelMain.setIcon(image);
        labelMain.setHorizontalAlignment(SwingConstants.CENTER); 
	  labelPanel.add(labelMain, BorderLayout.CENTER);
	  top.add(labelPanel, BorderLayout.SOUTH);

        // center panel -- split pane
        JSplitPane splitPane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT);
        pane.add(splitPane, BorderLayout.CENTER);
        splitPane.setContinuousLayout(true);
        splitPane.setOneTouchExpandable(true);
        splitPane.setDividerSize(3);
        
        JScrollPane splitPaneLeft = createTree();
        splitPane.setLeftComponent(splitPaneLeft);        
        splitPaneLeft.setMinimumSize(new Dimension(100, 100)); 

        goalHtmlPane = new JEditorPane();
        splitPane.setRightComponent(goalHtmlPane);
        goalHtmlPane.setMinimumSize(new Dimension(50, 50)); 
        displayHtmlFile("default.html");
   
        /*
        // bottom panel
        JPanel bottomPane = new JPanel();
	  bottomPane.setLayout(new BorderLayout());
        pane.add(bottomPane, BorderLayout.SOUTH);

        // bottom_1 panel
        JPanel bottom_1 = createHorizontalPanel();
        bottomPane.add(bottom_1, BorderLayout.WEST);
	  bottom_1.setBorder(new CompoundBorder(new TitledBorder(null, "Goals",
							  TitledBorder.LEFT, TitledBorder.TOP), border5));
	  bottom_1.setLayout(new FlowLayout(FlowLayout.LEFT));
        bottom_1.add(new JButton("Insert"));
        bottom_1.add(new JButton("Remove"));

        // bottom_2 panel
        JPanel bottom_2 = createHorizontalPanel();
        bottomPane.add(bottom_2, BorderLayout.CENTER);
	  bottom_2.setBorder(new CompoundBorder(new TitledBorder(null, "Plans",
							  TitledBorder.LEFT, TitledBorder.TOP), border5));
	  bottom_2.setLayout(new FlowLayout(FlowLayout.LEFT));
        bottom_2.add(new JButton("Insert"));
        bottom_2.add(new JButton("Remove"));
        
        bottomPane.add(new JButton("Exit"), BorderLayout.EAST); 
        */

        
        goalFrame.setJMenuBar(menuBar);
        goalFrame.setSize(400, 340);
        goalFrame.setVisible(true);

    } 

    public JScrollPane createTree() {

        DefaultMutableTreeNode goalTreeNode = new DefaultMutableTreeNode("Goal Tree");
        DefaultMutableTreeNode goal = null ;
	  DefaultMutableTreeNode subGoal = null;
	  DefaultMutableTreeNode sub2Goal = null; 
        DefaultMutableTreeNode currentGoal = null; 


        try {
            FileReader fileReader = new FileReader(GOAL_FILE);
            BufferedReader reader = new BufferedReader(fileReader);
           
	      // read one line at a time, put into tree
	      String line = reader.readLine();
	      while(line != null) {
		    // System.out.println("reading in: ->" + line + "<-");
		    char linetype = line.charAt(0);
		    switch(linetype) {
		        case 'G':
		            goal = new DefaultMutableTreeNode(line.substring(2));
            		goalTreeNode.add(goal);
                        currentGoal = goal;
   		            break;
		        case 'S':
		            if(goal != null) {
		                goal.add(subGoal = new DefaultMutableTreeNode(line.substring(2)));
                            currentGoal = subGoal;
		            }
		            break;
		        case 'U':
		            if(subGoal != null) {
		                subGoal.add(sub2Goal = new DefaultMutableTreeNode(line.substring(2)));
                            currentGoal = sub2Goal;
		            }
         		      break;
		        case 'P':
                        
		            // line.substring(2)
		            break;
		       default:
		            break;
		    }
		    line = reader.readLine();
            }      
	  } catch (IOException e) {}


        goalTree = new JTree(goalTreeNode) {
	      public Insets getInsets() {
		    return new Insets(5,5,5,5);
	      }
        }; 
        goalTree.getSelectionModel().setSelectionMode
                (TreeSelectionModel.SINGLE_TREE_SELECTION);
        goalTree.putClientProperty("JTree.lineStyle", "Angled");
        goalTree.addTreeSelectionListener(new GoalTreeSelectionListener());	  

        return new JScrollPane(goalTree);
    } 

    private void displayHtmlFile(String fileName) {
      
        URL url;
        String prefix = "file:" 
                        + System.getProperty("user.dir")
                        + System.getProperty("file.separator");
        try {
             url = new URL(prefix + fileName);
        } catch (java.net.MalformedURLException exc) {
             url = null;
             System.err.println("Attempted to create a "
                                   + "a bad URL: " + url);
        }

        try {
            goalHtmlPane.setPage(url);
        } catch (IOException e) {
            System.err.println("Attempted to read a bad URL: " + url);
        }
    }
 
    private class GoalTreeSelectionListener implements TreeSelectionListener {

        public void valueChanged(TreeSelectionEvent ev) {

            String selectedGoal = goalTree.getLastSelectedPathComponent().toString();
            
            System.out.println("Selected: tree -- " + selectedGoal);
           
            if (selectedGoal.equals("find buyer"))
                displayHtmlFile("findBuyer.html");
            else if (selectedGoal.equals("negotiate"))
                displayHtmlFile("negotiation.html");
            else if (selectedGoal.equals("sub goal 1"))
                displayHtmlFile("default.html");
            else
                displayHtmlFile("default.html");
           
        } // method valueChanged
    } // class GoalTreeSelectionListener 


    ///////////////// 
    // Goal Module //
    /////////////////

    public void editPlan() {
        
        editGoal();
    }
  

    //////////////////// 
    // Knowledge-base //
    ////////////////////

    
    public void editKnowledge() {

        frameKnowledge = new JFrame("AirTicketSeller agent knowledge-base (Part 1)");
        buttonActionListener = new ButtonActionListener();
        JMenuBar menuBar = new JMenuBar();

        EmptyBorder border5 = new EmptyBorder(5,5,5,5);
        EmptyBorder border5_0 = new EmptyBorder(5,0,0,0);
        EmptyBorder border10 = new EmptyBorder(10,10,10,10);

        JLabel labelMain = new JLabel("AirTicketSeller (ID: " 
                                                   + getAgentName() + ")");
        JLabel labelFrom = new JLabel("From (city)");
        JLabel labelDepartureTime = new JLabel("  Departure Time");
        JLabel labelTo = new JLabel("To (city)");
        JLabel labelArrivalTime = new JLabel("  Arrival Time");  
        JLabel labelDate = new JLabel("Date");
        JLabel labelAirline = new JLabel("  Airline");  
        JLabel labelStartingPrice = new JLabel("Starting Price $");
        JLabel labelLowestPrice = new JLabel("  Lowest Price $");
        JLabel labelSoldTickets = new JLabel("Sold Tickets");
        JLabel labelAvailableTickets = new JLabel("  Available Tickets");

        // JTextField inputFrom = new JTextField("Chicago");
        // JTextField inputTo = new JTextField("Dayton");
        // JTextField inputDepartureTime = new JTextField("0:00am");
        // JTextField inputArrivalTime = new JTextField("0:00am");
        // JTextField inputDate = new JTextField("12/05/01");
        // JTextField inputAirline = new JTextField("United Airline");
        // JTextField inputStartingPrice = new JTextField("125.00");
        // JTextField inputLowestPrice = new JTextField("100.00"); 
        // JTextField inputSoldTickets = new JTextField("10");
        // JTextField inputAvailableTickets = new JTextField("15"); 
 
        JButton buttonSave = new JButton("Save"); 
        buttonSave.setActionCommand("Save");
        buttonSave.addActionListener(buttonActionListener);
        JButton buttonClear = new JButton("Clear");
        buttonClear.setActionCommand("Clear"); 
        buttonClear.addActionListener(buttonActionListener);
        JButton buttonPrev = new JButton("Prev");
        buttonPrev.setActionCommand("Prev");
        buttonPrev.addActionListener(buttonActionListener);
        JButton buttonNext = new JButton("Next");
        buttonNext.setActionCommand("Next"); 
        buttonNext.addActionListener(buttonActionListener);
        JButton buttonExit = new JButton("Exit");
        buttonExit.setActionCommand("Exit"); 
        buttonExit.addActionListener(buttonActionListener);

        Container pane = frameKnowledge.getContentPane();
        pane.setLayout(new BorderLayout());
       
        // top panel
        JPanel top = new JPanel();
	  top.setLayout(new BorderLayout());
	  pane.add(top, BorderLayout.NORTH);

	  menuBar = createMenuBar();
	  top.add(menuBar, BorderLayout.NORTH);

        JPanel labelPanel = new JPanel();
	  labelPanel.setLayout(new BorderLayout());
        labelPanel.setBorder(border5_0);
        ImageIcon image = new ImageIcon("..\\images\\seller.gif",  "Seller.icon");
        labelMain.setIcon(image);
        labelMain.setHorizontalAlignment(SwingConstants.CENTER); 
	  labelPanel.add(labelMain, BorderLayout.CENTER);
	  top.add(labelPanel, BorderLayout.SOUTH);

        // center panel
        JPanel center = createHorizontalPanel();
        pane.add(center, BorderLayout.CENTER);
	  center.setBorder(new CompoundBorder(new TitledBorder(null, "Air Ticket Info",
							  TitledBorder.LEFT, TitledBorder.TOP), border5));
	  center.setLayout(new GridLayout(0,4));
        
 
        //Layout the labels in a panel
        JPanel labelPane_1 = new JPanel();
        labelPane_1.setLayout(new GridLayout(0, 1));
        labelPane_1.add(labelFrom);
        labelPane_1.add(labelTo);
        labelPane_1.add(labelDate);
        labelPane_1.add(labelStartingPrice);
        labelPane_1.add(labelSoldTickets);

        //Layout the text fields in a panel
        JPanel fieldPane_1 = new JPanel();
        fieldPane_1.setLayout(new GridLayout(0, 1));
        fieldPane_1.add(inputFrom);
        fieldPane_1.add(inputTo);
        fieldPane_1.add(inputDate);
        fieldPane_1.add(inputStartingPrice);
        fieldPane_1.add(inputSoldTickets);

        //Layout the labels in a panel
        JPanel labelPane_2 = new JPanel();
        labelPane_2.setLayout(new GridLayout(0, 1));
        labelPane_2.add(labelDepartureTime);
        labelPane_2.add(labelArrivalTime);
        labelPane_2.add(labelAirline); 
        labelPane_2.add(labelLowestPrice);
        labelPane_2.add(labelAvailableTickets);

        //Layout the text fields in a panel
        JPanel fieldPane_2 = new JPanel();
        fieldPane_2.setLayout(new GridLayout(0, 1));
        fieldPane_2.add(inputDepartureTime);
        fieldPane_2.add(inputArrivalTime);
        fieldPane_2.add(inputAirline);
        fieldPane_2.add(inputLowestPrice);
        fieldPane_2.add(inputAvailableTickets);


        //Put the panels in another panel, labels on left, text fields on right
        center.add(labelPane_1);
        center.add(fieldPane_1);
        center.add(labelPane_2);
        center.add(fieldPane_2);
        // bottom panel
        JPanel bottom = createHorizontalPanel();
        pane.add(bottom, BorderLayout.SOUTH);
	  bottom.setBorder(new CompoundBorder(new TitledBorder(null, "Buttons",
							  TitledBorder.LEFT, TitledBorder.TOP), border5));
	  bottom.setLayout(new FlowLayout(FlowLayout.LEFT));

	  bottom.add(buttonSave);
        bottom.add(buttonClear);
        bottom.add(buttonPrev);
        bottom.add(buttonNext);
        bottom.add(buttonExit);
        


        frameKnowledge.setJMenuBar(menuBar);
        frameKnowledge.setSize(400, 300);
        frameKnowledge.setVisible(true);

           
    }


    private JMenuBar createMenuBar() {

        JMenuBar menuBar = new JMenuBar();

        // File Menu, F - Mnemonic
        JMenu fileMenu = new JMenu("File");
        fileMenu.setMnemonic(KeyEvent.VK_F);
        menuBar.add(fileMenu);
   
        // Edit Menu, E - Mnemonic
        JMenu editMenu = new JMenu("Edit");
        editMenu.setMnemonic(KeyEvent.VK_E);
        menuBar.add(editMenu);
                 
        return menuBar;
    }

    public JPanel createHorizontalPanel() {
        JPanel p = new JPanel();
        p.setLayout(new BoxLayout(p, BoxLayout.X_AXIS));
        p.setAlignmentY(Component.TOP_ALIGNMENT);
        p.setAlignmentX(Component.LEFT_ALIGNMENT);
        return p;
    }
    
    protected void saveKnowledge() {
 
        fromCity = inputFrom.getText();
        departureTime = inputDepartureTime.getText();
        toCity = inputTo.getText();
        arrivalTime = inputArrivalTime.getText();
        dateFly = inputDate.getText();      
        airlineFly = inputAirline.getText();
        try {
            startingPrice = Float.parseFloat(inputStartingPrice.getText());
            lowestPrice = Float.parseFloat(inputLowestPrice.getText());
            soldTickets = Integer.parseInt(inputSoldTickets.getText());
            availableTickets = Integer.parseInt(inputAvailableTickets.getText());         
        } catch (Exception e) {
            JOptionPane.showMessageDialog(null, "Please input the correct values. "
                               + "And save again !");      
            return;
        }
      
     console.appendInfo("From: " + fromCity + "   To: " + toCity);
     console.appendInfo("StartingPrice: " + startingPrice + "   LowestPrice: " + lowestPrice);
     console.appendInfo("SoldTickets: " + soldTickets + "    AvailableTickets: " + availableTickets);

    }
 
    private class ButtonActionListener implements ActionListener {

        public void actionPerformed(ActionEvent actionEvent){

            System.out.println("Selected: "+actionEvent.getActionCommand());

            if ((actionEvent.getActionCommand()).equals("Exit")) {
                if (console.confirmWindow("Are you sure to EXIT knowledge-base?"))
                    frameKnowledge.dispose();
            } else if ((actionEvent.getActionCommand()).equals("Save")) {
                saveKnowledge();
            }          
        } // method actionPerformed
    } // class ButtonActionListener 



}


