/*
 * @(#)Message.java	1.00 11/20/01
 *
 * Copyright (c) 2001 by Concurrent Software System Lab (CSSL), Computer Science 
 * Department, The University of Illinois at Chicago (UIC). All Rights Reserved.
 * 
 * CSSL grants you ("Licensee") a non-exclusive, royalty free, license to use,
 * modify and redistribute this software in source and binary code form,
 * provided that i) this copyright notice and license appear on all copies of
 * the software; and ii) Licensee does not utilize the software in a manner
 * which is disparaging to CSSL.
 * 
 * This software is provided "AS IS," without a warranty of any kind. ALL
 * EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES, INCLUDING ANY
 * IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE OR
 * NON-INFRINGEMENT, ARE HEREBY EXCLUDED. CSSL AND ITS LICENSORS SHALL NOT BE
 * LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING
 * OR DISTRIBUTING THE SOFTWARE OR ITS DERIVATIVES. IN NO EVENT WILL CSSL OR ITS
 * LICENSORS BE LIABLE FOR ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT,
 * INDIRECT, SPECIAL, CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER
 * CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF THE USE OF
 * OR INABILITY TO USE SOFTWARE, EVEN IF CSSL HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES.
 * 
 * This software is not designed or intended for use in on-line control of
 * aircraft, air traffic, aircraft navigation or aircraft communications; or in
 * the design, construction, operation or maintenance of any nuclear
 * facility. Licensee represents and warrants that it will not use or
 * redistribute the Software for such purposes.
 */

package agent.generic;

import net.jini.core.lookup.ServiceID;

import java.lang.String;
import java.io.Serializable;

public class Message implements Serializable {

    public final static int INTERNAL = 0;
    public final static int EXTERNAL = 1;
    public final static int PRIVATE = 2;
    // public final static int PUBLIC = 3; 

    public final static int WRONG = -1;
    public final static int SAY_HI = 0;
    public final static int REPLY_HI = 1;

    public final static int WELCOME = 1000; 
    public final static int FIND_SELLER = 1001;
    public final static int SELLER_YES = 1002;
    public final static int SELLER_NO = 1003; 
    public final static int FIND_BUYER = 1004;
    public final static int BUYER_YES = 1005;
    public final static int BUYER_NO = 1006; 
    public final static int DEPART_ARRIVE_PLACE = 1007; 
    public final static int PLACE_YES = 1008;
    public final static int PLACE_NO = 1009;
    public final static int DATE_NUMOFTICKETS = 1010;
    public final static int DATE_NUMOFTICKETS_YES = 1011;
    public final static int DATE_NUMOFTICKETS_NO = 1012; 

    public final static int ASK_PRICE = 1100;
    public final static int REPLY_PRICE = 1101; 
    public final static int NEGOTIATE_PRICE = 1102;
    public final static int REPLY_LOWER_PRICE = 1103;
    public final static int COMPARE_PRICE = 1104;
    public final static int CONFIRM_PRICE = 1105;

    private ServiceID senderID;
    private ServiceID receiverID;
    private String content;
    private int type;
    private int tag;

    public String[] varString;
    public int[] varInt;
    public float[] varFloat;

    // =========== 
    // Constructor
    // ===========

    public Message(ServiceID senderID, String content) {
         this.senderID = senderID;
         this.content = content;
    }    

    public Message(ServiceID senderID, String content, int type) {
        this.senderID = senderID;
        this.content = content;
        this.type = type;
    } 

    // Functions
    
    public void setSenderID(ServiceID senderID) {
        this.senderID = senderID;
    }  
 
    public ServiceID getSenderID() {
        return this.senderID;
    }  

    public void setReceiverID(ServiceID receiverID) {
        this.receiverID = receiverID;
    }  
 
    public ServiceID getReceiverID() {
        return this.receiverID;
    } 

    public void setContent(String content) {
        this.content = content;
    }

    public String getContent() {
        return this.content;
    }       
     
    public void setTag(int theTag) {
       if (theTag < 0 && theTag > 3) return; 
       this.tag = theTag;
    }
        
    public int getTag() {
       return tag;
    }
        
    public void setType(int theType) { 
       this.type = theType;
    }
        
    public int getType() {
       return type;
    }

    public String toString() {
        return this.content;    
    } 

}
